@@ -6,7 +6,7 @@ from enum import Enum
 from functools import partial
 from inspect import Parameter, _empty, signature
 from pathlib import Path
-from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Tuple, Type, TypeVar, Union, cast
+from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Type, TypeVar, Union, cast
 
 try:
     from typing import get_args  # type: ignore
@@ -20,6 +20,7 @@ from flytekit.extras.sqlite3.task import SQLite3Task
 from sklearn.model_selection import train_test_split
 
 import unionml.type_guards as type_guards
+from unionml.defaults import DEFAULT_RESOURCES
 from unionml.utils import inner_task
 
 R = TypeVar("R")  # raw data
@@ -81,6 +82,7 @@ class Dataset(TrackedInstance):
         self._feature_transformer = self._default_feature_transformer
 
         self._reader = None
+        self._reader_task_kwargs: Optional[Dict[str, Any]] = None
         self._reader_input_types: Optional[List[Parameter]] = None
         self._dataset_datatype: Optional[Dict[str, Type]] = None
         self._dataset_task = None
@@ -102,7 +104,7 @@ class Dataset(TrackedInstance):
 
         type_guards.guard_reader(fn)
         self._reader = fn
-        self._reader_task_kwargs = reader_task_kwargs
+        self._reader_task_kwargs = {"requests": DEFAULT_RESOURCES, "limits": DEFAULT_RESOURCES, **reader_task_kwargs}
         return fn
 
     def loader(self, fn):
@@ -181,7 +183,9 @@ class Dataset(TrackedInstance):
 
         And it should return the data structure needed for model training.
         """
-        type_guards.guard_feature_loader(fn, self.parser_return_types[self._parser_feature_key])
+        # TODO: the feature_loader input should be compatible with the feature_reader (to be implemented) or
+        # pathlib.Path
+        type_guards.guard_feature_loader(fn, Any)
         self._feature_loader = fn
         return fn
 
@@ -194,7 +198,8 @@ class Dataset(TrackedInstance):
         2. When the FastAPI app `/predict/` endpoint is invoked with features passed in as JSON or string encoding.
         3. When the `model.predict` or `model.remote_predict` functions are invoked.
         """
-        type_guards.guard_feature_transformer(fn, self.parser_return_types[self._parser_feature_key])
+        sig = signature(self._feature_loader)
+        type_guards.guard_feature_transformer(fn, sig.return_annotation)
         self._feature_transformer = fn
         return fn
 
@@ -318,11 +323,17 @@ class Dataset(TrackedInstance):
         splits = self._splitter(data, **splitter_kwargs)
 
         if len(splits) == 1:
-            return {"train": self._parser(splits[0], **parser_kwargs)}
+            train_data = [*self._parser(splits[0], **parser_kwargs)]
+            train_data[self._parser_feature_key] = self._feature_transformer(train_data[self._parser_feature_key])
+            return {"train": train_data}
 
         train_split, test_split = splits
-        train_data = self._parser(train_split, **parser_kwargs)
-        test_data = self._parser(test_split, **parser_kwargs)
+        train_data = [*self._parser(train_split, **parser_kwargs)]
+        test_data = [*self._parser(test_split, **parser_kwargs)]
+
+        train_data[self._parser_feature_key] = self._feature_transformer(train_data[self._parser_feature_key])
+        test_data[self._parser_feature_key] = self._feature_transformer(test_data[self._parser_feature_key])
+
         return {
             "train": train_data,
             "test": test_data,
@@ -334,11 +345,8 @@ class Dataset(TrackedInstance):
         This function uses the following registered functions to create model-ready features:
 
         - :meth:`unionml.dataset.Dataset.feature_loader`
-        - :meth:`unionml.dataset.Dataset.parser`
         - :meth:`unionml.dataset.Dataset.feature_transformer`
         """
-        # parsed_data = self._parser(self._feature_loader(features), self._features, self._targets)
-        # features = parsed_data[self._parser_feature_key]
         features = self._feature_loader(features)
         return self._feature_transformer(features)
 
@@ -374,8 +382,8 @@ class Dataset(TrackedInstance):
         return ReaderReturnTypeSource.LOADER if self._loader != self._default_loader else ReaderReturnTypeSource.READER
 
     @property
-    def parser_return_types(self) -> Iterable[Type]:
-        """Get an iterable of types produces by the parser."""
+    def parser_return_types(self) -> Tuple[Any, ...]:
+        """Get an iterable of types produced by the parser."""
         sig = signature(self._parser)
         return get_args(sig.return_annotation)
 
@@ -385,12 +393,19 @@ class Dataset(TrackedInstance):
 
         The fallback behavior occurs if the user didn't define a ``feature_transformer`` function.
         """
-        parser_type = get_args(signature(self._parser).return_annotation)[self._parser_feature_key]
 
-        ft_type = signature(self._feature_transformer).return_annotation
-        if self._feature_transformer == self._default_feature_transformer:
-            # if the feature transformer is not user-defined, use the parser return signature
-            ft_type = get_args(signature(self._parser).return_annotation)[self._parser_feature_key]
+        parser_type = (
+            # use the reader/loader datatype if parser is the default parser, otherwise use parser return type
+            self.dataset_datatype["data"]
+            if self._parser == self._default_parser
+            else self.parser_return_types[self._parser_feature_key]
+        )
+
+        ft_type = (
+            signature(self._feature_loader).return_annotation
+            if self._feature_transformer == self._default_feature_transformer
+            else signature(self._feature_transformer).return_annotation
+        )
 
         if parser_type != ft_type:
             return cast(Type, Union[ft_type, parser_type])
@@ -406,7 +421,8 @@ class Dataset(TrackedInstance):
     ) -> "Dataset":
         dataset = cls(*args, **kwargs)
         dataset._dataset_task = task
-        dataset._dataset_datatype = task.python_interface.outputs
+        (_, dtype), *_ = task.python_interface.outputs.items()
+        dataset._dataset_datatype = {"data": dtype}
         dataset._reader_input_types = [
             Parameter(k, Parameter.KEYWORD_ONLY, annotation=v) for k, v in task.python_interface.inputs.items()
         ]
@@ -461,10 +477,10 @@ class Dataset(TrackedInstance):
 
     def _default_parser(
         self,
-        data: pd.DataFrame,
+        data: D,
         features: Optional[List[str]],
         targets: Optional[List[str]],
-    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
+    ) -> Tuple[D, D]:
         if not isinstance(data, pd.DataFrame):
             return (data,)  # type: ignore
 
