@@ -2,6 +2,7 @@
 
 import inspect
 import os
+import uuid
 from collections import OrderedDict
 from dataclasses import asdict, dataclass, field, is_dataclass, make_dataclass
 from functools import partial
@@ -9,6 +10,7 @@ from inspect import Parameter, signature
 from typing import IO, Any, Callable, Dict, List, NamedTuple, Optional, Tuple, Type, Union
 
 import joblib
+import pandas as pd
 import sklearn
 from dataclasses_json import dataclass_json
 from fastapi import FastAPI
@@ -17,9 +19,12 @@ from flytekit.configuration import Config
 from flytekit.core.tracker import TrackedInstance
 from flytekit.remote import FlyteRemote
 from flytekit.remote.executions import FlyteWorkflowExecution
+from flytekit.types.pickle import FlytePickle
 
 import unionml.type_guards as type_guards
+from unionml._logging import logger
 from unionml.dataset import Dataset
+from unionml.defaults import DEFAULT_RESOURCES
 from unionml.utils import inner_task, is_keras_model, is_pytorch_model
 
 
@@ -106,12 +111,14 @@ class Model(TrackedInstance):
         self._predict_from_features_task = None
 
         # user-provided task kwargs
-        self._train_task_kwargs = None
-        self._predict_task_kwargs = None
+        self._train_task_kwargs: Optional[Dict[str, Any]] = None
+        self._predict_task_kwargs: Optional[Dict[str, Any]] = None
 
         # dynamically defined types
         self._hyperparameter_type: Optional[Type] = None
 
+        self._patch_destination_dir: Optional[str] = None
+
     @property
     def artifact(self) -> Optional[ModelArtifact]:
         """Model artifact associated with the ``unionml.Model`` ."""
@@ -121,6 +128,11 @@ class Model(TrackedInstance):
     def artifact(self, new_value: ModelArtifact):
         self._artifact = new_value
 
+    @property
+    def dataset(self) -> Dataset:
+        """Exposes the ``unionml.Dataset`` associated with the model."""
+        return self._dataset
+
     @property
     def hyperparameter_type(self) -> Type:
         """Hyperparameter type of the model object based on the ``init`` function signature."""
@@ -183,29 +195,78 @@ class Model(TrackedInstance):
         self._init = fn
         return self._init
 
-    def trainer(self, fn=None, **train_task_kwargs):
-        """Register a function for training a model object."""
+    def trainer(self, fn: Callable = None, **train_task_kwargs):
+        """Register a function for training a model object.
+
+        This function is the primary entrypoint for defining your application's model-training behavior.
+
+        See the :ref:`User Guide <model_trainer>` for more.
+
+        :param fn: function to use as the trainer.
+        :param train_task_kwargs: keyword arguments to pass into the
+            :doc:`flytekit task <flytekit:generated/flytekit.task>` that will be composed of the input ``fn`` and
+            functions defined in the bound :py:class:`~unionml.dataset.Dataset`.
+        """
         if fn is None:
             return partial(self.trainer, **train_task_kwargs)
 
-        type_guards.guard_trainer(fn, self.model_type, self._dataset.parser_return_types)
+        if self.dataset._parser == self.dataset._default_parser:
+            # Use the reader/loader datatype if parser is the default parser, otherwise use parser return type.
+            # Additionally, special-case dataframes so that they are treated as two dataframes: one for features
+            # and another for targets.
+            expected_types = (
+                tuple([self.dataset.dataset_datatype["data"]] * 2)
+                if self.dataset.dataset_datatype["data"] is pd.DataFrame
+                else (self.dataset.dataset_datatype["data"],)
+            )
+        else:
+            expected_types = self.dataset.parser_return_types
+
+        type_guards.guard_trainer(fn, self.model_type, expected_types)
         self._trainer = fn
-        self._train_task_kwargs = train_task_kwargs
+        self._train_task_kwargs = {"requests": DEFAULT_RESOURCES, "limits": DEFAULT_RESOURCES, **train_task_kwargs}
         return self._trainer
 
     def predictor(self, fn=None, **predict_task_kwargs):
-        """Register a function that generates predictions from a model object."""
+        """Register a function that generates predictions from a model object.
+
+        This function is the primary entrypoint for defining your application's prediction behavior.
+
+        See the :ref:`User Guide <model_predictor>` for more.
+
+        :param fn: function to use as the predictor.
+        :param train_task_kwargs: keyword arguments to pass into the
+            :doc:`flytekit task <flytekit:generated/flytekit.task>` that will be composed of the input ``fn`` and
+            functions defined in the bound :py:class:`~unionml.dataset.Dataset`.
+        """
         if fn is None:
             return partial(self.predictor, **predict_task_kwargs)
 
         type_guards.guard_predictor(fn, self.model_type, self._dataset.feature_type)
         self._predictor = fn
-        self._predict_task_kwargs = predict_task_kwargs
+        self._predict_task_kwargs = {
+            "requests": DEFAULT_RESOURCES,
+            "limits": DEFAULT_RESOURCES,
+            **predict_task_kwargs,
+        }
         return self._predictor
 
     def evaluator(self, fn):
         """Register a function for producing metrics for given model object."""
-        type_guards.guard_evaluator(fn, self.model_type, self._dataset.parser_return_types)
+
+        if self.dataset._parser == self.dataset._default_parser:
+            # Use the reader/loader datatype if parser is the default parser, otherwise use parser return type.
+            # Additionally, special-case dataframes so that they are treated as two dataframes: one for features
+            # and another for targets.
+            expected_types = (
+                tuple([self.dataset.dataset_datatype["data"]] * 2)
+                if self.dataset.dataset_datatype["data"] is pd.DataFrame
+                else (self.dataset.dataset_datatype["data"],)
+            )
+        else:
+            expected_types = self.dataset.parser_return_types
+
+        type_guards.guard_evaluator(fn, self.model_type, expected_types)
         self._evaluator = fn
         return self._evaluator
 
@@ -261,11 +322,12 @@ class Model(TrackedInstance):
             dataset_task,
             **{k: wf.inputs[k] for k in dataset_task.python_interface.inputs},
         )
+        (_, dataset_promise), *_ = dataset_node.outputs.items()
         train_node = wf.add_entity(
             train_task,
             **{
                 hyperparam_arg: wf.inputs[hyperparam_arg],
-                **dataset_node.outputs,
+                **{"data": dataset_promise},
                 **{arg: wf.inputs[arg] for arg in trainer_param_types},
                 **{arg: wf.inputs[arg] for arg in ["loader_kwargs", "splitter_kwargs", "parser_kwargs"]},
             },
@@ -290,8 +352,9 @@ class Model(TrackedInstance):
             dataset_task,
             **{k: wf.inputs[k] for k in dataset_task.python_interface.inputs},
         )
+        (_, dataset_promise), *_ = dataset_node.outputs.items()
         predict_node = wf.add_entity(
-            predict_task, **{"model_object": wf.inputs["model_object"], **dataset_node.outputs}
+            predict_task, **{"model_object": wf.inputs["model_object"], **{"data": dataset_promise}}
         )
         for output_name, promise in predict_node.outputs.items():
             wf.add_workflow_output(output_name, promise)
@@ -403,8 +466,8 @@ class Model(TrackedInstance):
             **self._predict_task_kwargs,
         )
         def predict_task(model_object, **kwargs):
-            parsed_data = self._dataset._parser(kwargs[data_arg_name], **self._dataset.parser_kwargs)
-            features = parsed_data[self._dataset._parser_feature_key]
+            parsed_data = self.dataset._parser(kwargs[data_arg_name], **self._dataset.parser_kwargs)
+            features = self.dataset._feature_transformer(parsed_data[self._dataset._parser_feature_key])
             return self._predictor(model_object, features)
 
         self._predict_task = predict_task
@@ -451,15 +514,26 @@ class Model(TrackedInstance):
 
         :param hyperparameters: a dictionary mapping hyperparameter names to values. This is passed into the
             ``init`` callable to initialize a model object.
-        :param loader_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.loader` function.
+        :param loader_kwargs: key-word arguments to pass to the registered :meth:`unionml.Dataset.loader` function.
             This will override any defaults set in the function definition.
-        :param splitter_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.splitter` function.
+        :param splitter_kwargs: key-word arguments to pass to the registered :meth:`unionml.Dataset.splitter` function.
             This will override any defaults set in the function definition.
-        :param parser_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.parser` function.
+        :param parser_kwargs: key-word arguments to pass to the registered :meth:`unionml.Dataset.parser` function.
             This will override any defaults set in the function definition.
         :param trainer_kwargs: a dictionary mapping training parameter names to values. There training parameters
             are determined by the keyword-only arguments of the ``model.trainer`` function.
         :param reader_kwargs: keyword arguments that correspond to the :meth:`unionml.Dataset.reader` method signature.
+
+        The train method invokes an execution graph that composes together the following functions to train and evaluate a model:
+
+            - :meth:`unionml.Dataset.reader`
+            - :meth:`unionml.Dataset.loader`
+            - :meth:`unionml.Dataset.splitter`
+            - :meth:`unionml.Dataset.parser`
+            - :meth:`unionml.Model.trainer`
+            - :meth:`unionml.Model.predictor`
+            - :meth:`unionml.Model.evaluator`
+
         """
         trainer_kwargs = {} if trainer_kwargs is None else trainer_kwargs
         model_obj, hyperparameters, metrics = self.train_workflow()(
@@ -510,23 +584,50 @@ class Model(TrackedInstance):
         return self._saver(self.artifact.model_object, self.artifact.hyperparameters, file, *args, **kwargs)
 
     def load(self, file: Union[str, os.PathLike, IO], *args, **kwargs):
-        """Load a model object from disk."""
-        return self._loader(file, *args, **kwargs)
+        """Load a model object from disk.
 
-    def serve(self, app: FastAPI, remote: bool = False, model_version: str = "latest"):
+        :file: a string or path-like object to load a model from.
+        :param args: positional arguments forwarded to :meth:`unionml.Model.loader` .
+        :param kwargs: key-word arguments forwarded to :meth:`unionml.Model.loader` .
+        """
+        self.artifact = ModelArtifact(self._loader(file, *args, **kwargs))
+        return self.artifact.model_object
+
+    def load_from_env(self, env_var: str = "UNIONML_MODEL_PATH", *args, **kwargs):
+        """Load a model object from an environment variable pointing to the model file.
+
+        :env_var: environment variable referencing a path to load a model from.
+        :param args: positional arguments forwarded to :meth:`unionml.Model.loader` .
+        :param kwargs: key-word arguments forwarded to :meth:`unionml.Model.loader` .
+        """
+        model_path = os.getenv(env_var)
+        if model_path is None:
+            raise ValueError("env_var for model path {env_var} doesn't exist.")
+        return self.load(model_path)
+        self.artifact = ModelArtifact(self.load(model_path))
+        return self.artifact.model_object
+
+    def serve(
+        self,
+        app: FastAPI,
+        remote: bool = False,
+        app_version: Optional[str] = None,
+        model_version: str = "latest",
+    ):
         """Create a FastAPI serving app.
 
         :param app: A ``FastAPI`` app to use for model serving.
         """
         from unionml.fastapi import serving_app
 
-        serving_app(self, app, remote=remote, model_version=model_version)
+        serving_app(self, app, remote=remote, app_version=app_version, model_version=model_version)
 
     def remote(
         self,
         registry: Optional[str] = None,
         image_name: str = None,
         dockerfile: str = "Dockerfile",
+        patch_destination_dir: str = "/root",
         config_file: Optional[str] = None,
         project: Optional[str] = None,
         domain: Optional[str] = None,
@@ -536,10 +637,13 @@ class Model(TrackedInstance):
         :param registry: Docker registry used to push UnionML app.
         :param image_name: image name to give to the Docker image associated with the UnionML app.
         :param dockerfile: path to the Dockerfile used to package the UnionML app.
+        :param patch_destination_dir: path where the UnionML source is installed within the docker image, and in case of
+                                    patch registration, this would be replaced. In Flyte terms, patch registration is
+                                    often called fast-registration.
         :param config_file: path to the `flytectl config <https://docs.flyte.org/projects/flytectl/en/latest/>`__ to use for
             deploying your UnionML app to a Flyte backend.
         :param project: deploy your app to this Flyte project name.
-        :param project: deploy your app to this Flyte domain name.
+        :param domain: deploy your app to this Flyte domain name.
         """
         self._config_file = config_file
         self._registry = registry
@@ -547,6 +651,7 @@ class Model(TrackedInstance):
         self._dockerfile = dockerfile
         self._project = project
         self._domain = domain
+        self._patch_destination_dir = patch_destination_dir
 
     @property
     def _remote(self) -> Optional[FlyteRemote]:
@@ -564,33 +669,65 @@ class Model(TrackedInstance):
         )
         return self.__remote__
 
-    def remote_deploy(self, app_version: str = None):
-        """Deploy model services to a Flyte backend."""
+    def remote_deploy(self, app_version: str = None, allow_uncommitted: bool = False, patch: bool = False) -> str:
+        """
+        Deploy model services to a Flyte backend.
+
+        :param app_version: the version to use to deploy the UnionML app.
+        :param allow_uncommitted: If True, deploys uncommitted changes in the unionml project. Otherwise, raise a
+            :py:class`~unionml.remote.VersionFetchError`
+        :param patch: if True, this bypasses the Docker build process and only updates the UnionML app source code using
+            the latest available image.
+        :returns: app version string
+        """
+        from unionml import remote
+
         if self._remote is None:
             raise RuntimeError("First configure the remote client with the `Model.remote` method")
 
-        from unionml import remote
-
-        app_version = app_version or remote.get_app_version(allow_uncommitted=True)
-        image = remote.get_image_fqn(self, app_version, self._image_name)
-
-        os.environ["FLYTE_INTERNAL_IMAGE"] = image or ""
         _remote = self._remote
+        if not _remote.config.platform.endpoint.startswith("localhost") and self.registry is None:
+            raise ValueError("You must specify a registry in `model.remote` when deploying to a remote cluster.")
 
+        # if user wants to deploy by patching, then we implicitly want to allow for uncommitted changes.
+        is_explicit_app_version = app_version is not None
+        app_version = app_version or remote.get_app_version(allow_uncommitted=allow_uncommitted or patch)
+        image = remote.get_image_fqn(self, app_version, self._image_name)
+        os.environ["FLYTE_INTERNAL_IMAGE"] = image or ""
+
+        logger.info(f"Using remote endpoint {_remote.config.platform.endpoint}")
         remote.create_project(_remote, self._project)
-        if _remote.config.platform.endpoint.startswith("localhost"):
-            # assume that a localhost flyte_admin_url means that we want to use Flyte sandbox
-            remote.sandbox_docker_build(self, image)
+        if patch:
+            app_version = app_version if is_explicit_app_version else f"{app_version}-patch{uuid.uuid4().hex[:7]}"
         else:
-            remote.docker_build_push(self, image)
+            logger.info(
+                "Building docker image. If no dependencies have changed, use the `patch` arg to bypass "
+                "docker build process"
+            )
+            if _remote.config.platform.endpoint.startswith("localhost"):
+                # assume that a localhost flyte_admin_url means that we want to use Flyte sandbox
+                remote.sandbox_docker_build(self, image)
+            else:
+                remote.docker_build_push(self, image)
 
-        args = [_remote._default_project, _remote._default_domain, app_version]
+        logger.info(f"Deploying app version {app_version}")
         for wf in [
             self.train_workflow(),
             self.predict_workflow(),
             self.predict_from_features_workflow(),
         ]:
-            remote.deploy_wf(wf, _remote, image, *args)
+            remote.deploy_wf(
+                wf,
+                _remote,
+                image,
+                project=_remote.default_project,
+                domain=_remote.default_domain,
+                version=app_version,
+                patch=patch,
+                patch_destination_dir=self._patch_destination_dir,
+            )
+
+        return app_version
 
     def remote_train(
         self,
@@ -607,27 +744,25 @@ class Model(TrackedInstance):
         """Train a model object on a remote Flyte backend.
 
         :param app_version: if provided, executes a training job using the specified UnionML app version. By default,
-            this uses the current git sha of the repo, which versions your UnionML app.
+            this uses the latest app version deployed on the Flyte remote cluster..
         :param wait: if True, this is a synchronous operation, returning a ``ModelArtifact``. Otherwise, this
             function returns a ``FlyteWorkflowExecution``.
         :param hyperparameters: a dictionary mapping hyperparameter names to values. This is passed into the
             ``init`` callable to initialize a model object.
-        :param loader_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.loader` function.
+        :param loader_kwargs: key-word arguments to pass to the registered :meth:`unionml.Dataset.loader` function.
             This will override any defaults set in the function definition.
-        :param splitter_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.splitter` function.
+        :param splitter_kwargs: key-word arguments to pass to the registered :meth:`unionml.Dataset.splitter` function.
             This will override any defaults set in the function definition.
-        :param parser_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.parser` function.
+        :param parser_kwargs: key-word arguments to pass to the registered :meth:`unionml.Dataset.parser` function.
             This will override any defaults set in the function definition.
         :param trainer_kwargs: a dictionary mapping training parameter names to values. There training parameters
             are determined by the keyword-only arguments of the ``model.trainer`` function.
         :param reader_kwargs: keyword arguments that correspond to the :meth:`unionml.Dataset.reader` method signature.
+        :returns: the trained model if wait is ``True``, or a ``FlyteWorkflowExecution`` object if wait is ``False``.
         """
         if self._remote is None:
             raise RuntimeError("First configure the remote client with the `Model.remote` method")
 
-        from unionml import remote
-
-        app_version = app_version or remote.get_app_version(allow_uncommitted=True)
         train_wf = self._remote.fetch_workflow(name=self.train_workflow_name, version=app_version)
         execution = self._remote.execute(
             train_wf,
@@ -640,7 +775,6 @@ class Model(TrackedInstance):
             },
             project=self._remote.default_project,
             domain=self._remote.default_domain,
-            wait=wait,
             type_hints={
                 "hyperparameters": self.hyperparameter_type,
                 "loader_kwargs": self._dataset.loader_kwargs_type,
@@ -649,13 +783,15 @@ class Model(TrackedInstance):
             },
         )
         console_url = self._remote.generate_console_url(execution)
-        print(
+        logger.info(
             f"Executing {train_wf.id.name}, execution name: {execution.id.name}."
             f"\nGo to {console_url} to see the execution in the console."
         )
+
         if not wait:
             return execution
 
+        self.remote_wait(execution)
         self.remote_load(execution)
         return self.artifact
 
@@ -674,7 +810,7 @@ class Model(TrackedInstance):
         arguments that will be forwarded to the :meth:`unionml.Dataset.reader` method as the feature source.
 
         :param app_version: if provided, executes a prediction job using the specified UnionML app version. By default,
-            this uses the current git sha of the repo, which versions your UnionML app.
+            this uses the latest app version deployed on the Flyte remote cluster.
         :param model_version: if provided, executes a prediction job using the specified model version. By default, this
             uses the latest Flyte execution id as the model version.
         :param wait: if True, this is a synchronous operation, returning a ``ModelArtifact``. Otherwise, this
@@ -685,18 +821,14 @@ class Model(TrackedInstance):
             - :meth:`unionml.dataset.Dataset.feature_loader`
             - :meth:`unionml.dataset.Dataset.feature_transformer`
         :param reader_kwargs: keyword arguments that correspond to the :meth:`unionml.Dataset.reader` method signature.
+        :returns: the predictions if wait is ``True``, or a ``FlyteWorkflowExecution`` object if wait is ``False``.
         """
         if self._remote is None:
             raise RuntimeError("First configure the remote client with the `Model.remote` method")
 
         from unionml import remote
 
-        app_version = app_version or remote.get_app_version(allow_uncommitted=True)
         model_artifact = remote.get_model_artifact(self, app_version, model_version)
-
-        if (features is not None and len(reader_kwargs) > 0) or (features is None and len(reader_kwargs) == 0):
-            raise ValueError("You must provide only one of `features` or `reader_kwargs`")
-
         inputs = {"model_object": model_artifact.model_object}
         if features is None:
             workflow_name = self.predict_workflow_name
@@ -722,7 +854,7 @@ class Model(TrackedInstance):
             type_hints=type_hints,
         )
         console_url = self._remote.generate_console_url(execution)
-        print(
+        logger.info(
             f"Executing {predict_wf.id.name}, execution name: {execution.id.name}."
             f"\nGo to {console_url} to see the execution in the console."
         )
@@ -745,13 +877,18 @@ class Model(TrackedInstance):
         if self._remote is None:
             raise ValueError("You must call `model.remote` to attach a remote backend to this model.")
         if not execution.is_done:
-            print(f"Waiting for execution {execution.id.name} to complete...")
+            logger.info(f"Waiting for execution {execution.id.name} to complete...")
             execution = self.remote_wait(execution)
-            print("Done.")
+            logger.info("Done.")
 
         with self._remote.remote_context():
+            try:
+                model_object = execution.outputs.get("model_object", as_type=self.model_type)
+            except ValueError:
+                model_object = execution.outputs.get("model_object", as_type=FlytePickle)
+
             self.artifact = ModelArtifact(
-                execution.outputs["model_object"],
+                model_object,
                 execution.outputs["hyperparameters"],
                 execution.outputs["metrics"],
             )
@@ -834,7 +971,11 @@ class Model(TrackedInstance):
             import torch
 
             deserialized_model = torch.load(file, *args, **kwargs)
-            model = model_type(**deserialized_model["hyperparameters"])
+            if self._init_callable is not None:
+                model = self._init(deserialized_model["hyperparameters"])
+            else:
+                model = model_type(**deserialized_model["hyperparameters"])
+
             model.load_state_dict(deserialized_model["model_obj"])
             return model
         elif is_keras_model(model_type):
